package ch.unige.solidify.controller;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.hateoas.RepresentationModel;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.exception.SolidifyUndeletableException;
import ch.unige.solidify.exception.SolidifyUnmodifiableException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.rest.Tool;
import ch.unige.solidify.service.CompositeResourceService;

public abstract class ResourceController<T extends Resource<T>> extends SolidifyController
        implements ControllerWithHateoasHome {

  @Autowired
  protected CompositeResourceService<T> itemService;

  @GetMapping
  public HttpEntity<Collection<T>> list(@ModelAttribute T search, Pageable pageable) {

    Specification<T> spec = search.getSpecification();
    final Page<T> listItem = this.itemService.findAll(spec, pageable);
    for (final T t : listItem) {
      addLinks(t);
    }
    Collection<T> collection = new Collection<>(listItem, pageable);
    collection.add(linkTo(this.getClass()).withSelfRel());
    collection.add(Tool.parentLink((linkTo(this.getClass())).toUriComponentsBuilder())
            .withRel(ActionName.MODULE));
    this.addSortLinks(linkTo(this.getClass()), collection);
    this.addPageLinks(linkTo(this.getClass()), collection, pageable);
    addOthersLinks(collection);
    return new ResponseEntity<>(collection, HttpStatus.OK);
  }

  @PostMapping
  public HttpEntity<T> create(@RequestBody T t) {

    if (t.getResId() != null && this.itemService.existsById(t.getResId())) {
      throw new SolidifyRuntimeException(this.messageService.get("validation.resource.alreadyexist",
              new Object[] { t.getResId() }));
    }

    this.setAuthenticatedUserProperties(t);

    T item = this.itemService.save(t);
    this.addLinks(item);
    return new ResponseEntity<>(item, HttpStatus.CREATED);
  }

  @GetMapping(value = SolidifyConstants.URL_ID)
  public HttpEntity<T> get(@PathVariable String id) {
    T t = this.itemService.findOne(id);
    this.addLinks(t);
    return new ResponseEntity<>(t, HttpStatus.OK);
  }

  @SuppressWarnings("squid:S3752")
  @RequestMapping(value = SolidifyConstants.URL_ID, method = { RequestMethod.PUT,
          RequestMethod.PATCH })
  public HttpEntity<T> update(@PathVariable String id, @RequestBody T t2) {

    T t1 = this.itemService.findOne(id);

    if (!t1.isModifiable()) {
      throw new SolidifyUnmodifiableException(
              this.messageService.get("validation.resource.unmodifiable",
                      new Object[] { t1.getClass().getSimpleName() + ": " + id }));
    }

    if (t1.update(t2)) {
      this.setAuthenticatedUserProperties(t1);
      T item = this.itemService.save(t1);
      this.addLinks(item);
      return new ResponseEntity<>(item, HttpStatus.OK);
    } else {
      return new ResponseEntity<>(t1, HttpStatus.OK);
    }
  }

  @DeleteMapping(value = SolidifyConstants.URL_ID)
  public HttpStatus delete(@PathVariable String id) {
    T t = this.itemService.findOne(id);

    if (!t.isDeletable()) {
      throw new SolidifyUndeletableException(
              this.messageService.get("validation.resource.undeletable",
                      new Object[] { t.getClass().getSimpleName() + ": " + id }));
    }

    this.itemService.delete(id);
    return HttpStatus.OK;
  }

  protected void addLinks(T t) {
    t.removeLinks();
    t.addLinks(linkTo(this.getClass()), true, true);
  }

  protected <W extends RepresentationModel<W>> void addOthersLinks(W w) {
    // Do nothing
    // Could be overridden by subclass
  }

}
