package ch.unige.solidify.controller;

import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Paths;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.InputStreamResource;
import org.springframework.data.domain.Pageable;
import org.springframework.hateoas.server.mvc.WebMvcLinkBuilder;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.authentication.AnonymousAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.web.multipart.MultipartFile;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.ResourceBase;
import ch.unige.solidify.rest.Tool;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;

public abstract class SolidifyController {
  private static final Logger log = LoggerFactory.getLogger(SolidifyController.class);

  @Autowired
  protected MessageService messageService;

  protected ResponseEntity<InputStreamResource> download(InputStream in, String fileTarget, String fileContentType, long fileSize) {
    if (StringTool.isNullOrEmpty(fileTarget) || StringTool.isNullOrEmpty(fileContentType)) {
      return new ResponseEntity<>(HttpStatus.SERVICE_UNAVAILABLE);
    }
    HttpHeaders respHeaders = new HttpHeaders();
    respHeaders.setContentType(MediaType.valueOf(fileContentType));
    if (fileSize > 0) {
      respHeaders.setContentLength(fileSize);
    }
    respHeaders.setContentDispositionFormData("attachment", fileTarget);

    return new ResponseEntity<>(new InputStreamResource(in), respHeaders, HttpStatus.OK);
  }

  protected boolean uploadData(MultipartFile file, String destinationFolder) {
    try {
      Files.copy(file.getInputStream(),
              Paths.get(destinationFolder).resolve(file.getOriginalFilename()));
    } catch (IOException e) {
      String message = "Error uploading file " + file;
      log.error(message, e);
      throw new SolidifyRuntimeException(message, e);
    }
    return true;
  }

  protected <T> void addSortLinks(WebMvcLinkBuilder linkBuilder, Collection<T> collection) {
    boolean creationExisted = true;
    boolean lastUpdateExisted = true;
    if (collection != null) {
      if (collection.get_data() != null) {
        if (collection.get_data().iterator().hasNext()) {
          T t = collection.get_data().iterator().next();
          if (t instanceof ResourceBase) {
            ResourceBase res = (ResourceBase) t;
            if (res.getCreation() == null)
              creationExisted = false;
            if (res.getLastUpdate() == null)
              lastUpdateExisted = false;
          }
        }
      }
      if (creationExisted) {
        collection.add(Tool.sort(linkBuilder.toUriComponentsBuilder(), ActionName.CREATION)
                .withRel(ActionName.LASTCREATED));
      }
      if (lastUpdateExisted) {
        collection.add(Tool.sort(linkBuilder.toUriComponentsBuilder(), ActionName.UPDATED)
                .withRel(ActionName.LASTUPDATE));
      }
    }
  }

  protected <T> void addPageLinks(WebMvcLinkBuilder linkBuilder, Collection<T> collection,
          Pageable pageable) {
    if (collection.get_page().hasPrevious()) {
      collection.add(
              Tool.pageLink(linkBuilder.toUriComponentsBuilder(), pageable.previousOrFirst().getPageSize(),
                      pageable.previousOrFirst().getPageNumber()).withRel(ActionName.PREVIOUS));
    }
    if (collection.get_page().hasNext()) {
      collection.add(Tool.pageLink(linkBuilder.toUriComponentsBuilder(), pageable.next().getPageSize(),
              pageable.next().getPageNumber()).withRel(ActionName.NEXT));
    }
  }

  protected ResourceBase setAuthenticatedUserProperties(ResourceBase resource) {
    String userId = this.getAuthenticatedUserId();

    if (userId != null) {
      resource.setUpdatedBy(userId);

      if (StringTool.isNullOrEmpty(resource.getCreatedBy())) {
        resource.setCreatedBy(userId);
      }
    } else {
      resource.setUpdatedBy(SolidifyConstants.NO_CREDENTIAL);
      if (StringTool.isNullOrEmpty(resource.getCreatedBy())) {
        resource.setCreatedBy(SolidifyConstants.NO_CREDENTIAL);
      }
    }

    return resource;
  }

  protected String getAuthenticatedUserId() {
    Authentication authentication = SecurityContextHolder.getContext().getAuthentication();

    if (authentication != null && !(authentication instanceof AnonymousAuthenticationToken)) {
      return authentication.getName();
    } else {
      return null;
    }
  }
}
