package ch.unige.solidify.controller;

import java.util.NoSuchElementException;
import javax.servlet.http.HttpServletRequest;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.web.servlet.error.ErrorAttributes;
import org.springframework.dao.CannotAcquireLockException;
import org.springframework.dao.ConcurrencyFailureException;
import org.springframework.dao.DataIntegrityViolationException;
import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.orm.ObjectOptimisticLockingFailureException;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.validation.BindException;
import org.springframework.validation.BindingResult;
import org.springframework.web.HttpRequestMethodNotSupportedException;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.annotation.ControllerAdvice;
import org.springframework.web.bind.annotation.ExceptionHandler;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRestException;
import ch.unige.solidify.exception.SolidifyUndeletableException;
import ch.unige.solidify.exception.SolidifyUnmodifiableException;
import ch.unige.solidify.exception.SolidifyValidationException;
import ch.unige.solidify.repository.SolidifyDbLockError;
import ch.unige.solidify.repository.SolidifyError;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.validation.ValidationError;

@ControllerAdvice
public class SolidifyControllerAdvice {

  private static final Logger log = LoggerFactory.getLogger(SolidifyControllerAdvice.class);
  private static final String ERROR_CONTENT_TYPE = "application/hal+json";

  private final ErrorAttributes errorAttributes;
  private final MessageService messageService;
  private final HttpHeaders headers = new HttpHeaders();

  public SolidifyControllerAdvice(ErrorAttributes errorAttributes, MessageService messageService) {
    headers.add(HttpHeaders.CONTENT_TYPE, ERROR_CONTENT_TYPE);
    this.errorAttributes = errorAttributes;
    this.messageService = messageService;
  }

  /******************************/

  /**
   * Used to catch Spring constraints validation errors (rules defined as annotations on entities)
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler(value = MethodArgumentNotValidException.class)
  public ResponseEntity<SolidifyError> handleException(MethodArgumentNotValidException exception, HttpServletRequest request) {
    BindingResult errors = exception.getBindingResult();
    ValidationError validationError = new ValidationError(errors);
    return buildBadRequestResponse(exception, request, validationError);
  }

  /**
   * Used to catch ValidationError set at the service level (such as linked resources not found)
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler(value = SolidifyValidationException.class)
  public ResponseEntity<SolidifyError> handleException(SolidifyValidationException exception, HttpServletRequest request) {
    return buildBadRequestResponse(exception, request, exception.getValidationError());
  }

  /**
   * Used to catch DataIntegrityViolationException
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler(value = DataIntegrityViolationException.class)
  public ResponseEntity<SolidifyError> handleException(DataIntegrityViolationException exception, HttpServletRequest request) {
    SolidifyError solidifyError = new SolidifyError(errorAttributes, request);
    solidifyError.setStatus(HttpStatus.BAD_REQUEST);
    solidifyError.setError("Data integrity violation exception");
    String exceptionMessage = exception.getMostSpecificCause().getMessage();
    if (exceptionMessage.contains("\n")) {
      exceptionMessage = exceptionMessage.substring(0, exceptionMessage.indexOf('\n'));
    }
    solidifyError.setMessage(exceptionMessage);
    return buildBadRequestResponse(exception, request, solidifyError);
  }

  /**
   * Used to catch: AccessDeniedException.class (not enough permissions),
   * SolidifyUnmodifiableException (resource unmodifiable), SolidifyUndeletableException (resource
   * undeletable)
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler(value = { AccessDeniedException.class, SolidifyUnmodifiableException.class, SolidifyUndeletableException.class })
  public ResponseEntity<SolidifyError> handleForbiddenException(Exception exception, HttpServletRequest request) {
    return buildResponse(exception, request, messageService.get("error.request.forbidden"), HttpStatus.FORBIDDEN);
  }

  /**
   * Used to catch requests errors (such as invalid JSON received)
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler(value = HttpMessageNotReadableException.class)
  public ResponseEntity<SolidifyError> handleException(HttpMessageNotReadableException exception, HttpServletRequest request) {

    logException(exception, request);

    String message = messageService.get("error.request.notreadable");

    Throwable nestedException = exception.getCause();
    if (nestedException instanceof JsonParseException) {
      message = messageService.get("error.request.json.malformed");
    }

    SolidifyError error = new SolidifyError(errorAttributes, request);
    error.setMessage(message);
    error.setError(exception.getMessage());
    error.setStatus(HttpStatus.BAD_REQUEST);

    return ResponseEntity.badRequest().headers(headers).body(error);
  }

  /**
   * Used to catch NoSuchElementException/EmptyResultDataAccessException
   *
   * @param exception
   * @param request
   * @return
   */
  @ExceptionHandler(value = { NoSuchElementException.class, EmptyResultDataAccessException.class, SolidifyResourceNotFoundException.class })
  public ResponseEntity<SolidifyError> handleNotFoundException(Exception exception, HttpServletRequest request) {
    return buildResponse(exception, request, messageService.get("error.request.notfound"), HttpStatus.NOT_FOUND);
  }

  @ExceptionHandler(value = IllegalStateException.class)
  public ResponseEntity<SolidifyError> handleException(IllegalStateException exception, HttpServletRequest request) {
    return buildResponse(exception, request, messageService.get("error.illegal.state"), HttpStatus.BAD_REQUEST);
  }

  @ExceptionHandler(value = BindException.class)
  public ResponseEntity<SolidifyError> handleException(BindException exception, HttpServletRequest request) {
    return buildResponse(exception, request, messageService.get("error.illegal.parameter"), HttpStatus.BAD_REQUEST);
  }

  @ExceptionHandler(value = HttpRequestMethodNotSupportedException.class)
  public ResponseEntity<SolidifyError> handleException(HttpRequestMethodNotSupportedException exception, HttpServletRequest request) {
    return buildResponse(exception, request, messageService.get("error.illegal.http.method"), HttpStatus.METHOD_NOT_ALLOWED);
  }

  // Exception handler for Optimistic lock or pessimistic lock
  @ExceptionHandler({ CannotAcquireLockException.class, ObjectOptimisticLockingFailureException.class })
  public ResponseEntity<SolidifyError> handleLockException(ConcurrencyFailureException exception, HttpServletRequest request) {
    logException(exception, request);
    SolidifyError error = new SolidifyDbLockError(errorAttributes, request, exception);
    return ResponseEntity.status(error.getStatus()).headers(headers).body(error);
  }

  // Exception handler for Solidify Rest client HttpClientErrorException
  @ExceptionHandler(value = SolidifyRestException.class)
  public ResponseEntity<SolidifyError> handleLockException(SolidifyRestException exception, HttpServletRequest request) {
    logException(exception, request);
    SolidifyError error = exception.getSourceError();
    if (error != null) {
      return ResponseEntity.status(error.getStatus()).headers(headers).body(error);
    } else {
      error = new SolidifyError(errorAttributes, request);
      return ResponseEntity.status(error.getStatus()).headers(headers).body(error);
    }
  }

  /**
   * Used to catch all exceptions not specifically catched by another ExceptionHandler
   *
   * @param exception
   * @param request
   * @returnlogException(exception, request);
   */
  @ExceptionHandler(value = Exception.class)
  public ResponseEntity<SolidifyError> handleException(Exception exception, HttpServletRequest request) {

    logException(exception, request);
    SolidifyError error = new SolidifyError(errorAttributes, request);
    return ResponseEntity.status(error.getStatus()).headers(headers).body(error);
  }

  private ResponseEntity<SolidifyError> buildResponse(Exception exception,
          HttpServletRequest request, String message, HttpStatus httpStatus) {
    logException(exception, request);

    SolidifyError error = new SolidifyError(errorAttributes, request);
    error.setMessage(message);
    error.setError(exception.getMessage());
    error.setStatus(httpStatus);

    return ResponseEntity.status(httpStatus).headers(headers).body(error);
  }

  private ResponseEntity<SolidifyError> buildBadRequestResponse(Exception exception, HttpServletRequest request,
          ValidationError validationError) {

    SolidifyError solidifyError = new SolidifyError(errorAttributes, request);
    solidifyError.setValidationError(validationError);
    solidifyError.setStatus(HttpStatus.BAD_REQUEST);

    return buildBadRequestResponse(exception, request, solidifyError);
  }

  private ResponseEntity<SolidifyError> buildBadRequestResponse(Exception exception, HttpServletRequest request, SolidifyError solidifyError) {
    logException(exception, request);
    logErrorAsJson(solidifyError);
    return ResponseEntity.badRequest().headers(headers).body(solidifyError);
  }

  private void logException(Exception exception, HttpServletRequest request) {
    log.error("Request url {} has generated an exception catched by controller advice: {}", request.getRequestURL(), exception);
    log.debug("Stacktrace", exception);

  }

  private void logErrorAsJson(Object errorObject) {
    try {
      ObjectMapper mapper = new ObjectMapper();
      log.error(mapper.writeValueAsString(errorObject));
    } catch (JsonProcessingException e) {
      log.error("unable to log validation error: {}", e);
    }
  }

}
