package ch.unige.solidify.service;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.BindingResult;
import org.springframework.validation.Validator;

import ch.unige.solidify.config.SolidifyProperties;
import ch.unige.solidify.exception.SolidifyValidationException;
import ch.unige.solidify.repository.SolidifyRepository;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.rest.SubResourceContainer;
import ch.unige.solidify.util.SearchCriteria;
import ch.unige.solidify.util.SearchOperation;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.validation.ValidationError;

/**
 * A service for composite resources
 * 
 * @param <T>
 *          the type of the resource
 */
public abstract class CompositeResourceService<T extends Resource<T>> {

  private static final Logger log = LoggerFactory.getLogger(CompositeResourceService.class);

  @Autowired
  protected MessageService messageService;

  @Autowired
  protected SolidifyRepository<T> itemRepository;

  @Autowired
  protected ResourceService<T> itemRestService;

  @Autowired
  private Validator validator;

  /**
   * Use a LinkedRESTResourceChecker instead of calling directly validateLinkedRESTResources() method
   * in order to be able to disable validation during tests. (during tests
   * TestLinkedRESTResourceChecker is injected, thus bypassing linked resources validation)
   */
  @Autowired
  private LinkedResourceChecker<T> linkedResourceValidator;

  @Autowired
  protected SolidifyProperties config;

  /************************************************
   * JPA
   ************************************************/

  protected boolean localExistsById(String resId) {
    return this.itemRepository.existsById(resId);
  }

  protected T localFindOne(String id) {
    return this.itemRepository.findById(id)
            .orElseThrow(() -> new NoSuchElementException("No resource with id " + id));
  }

  protected Page<T> localFindAll(Specification<T> spec, Pageable pageable) {
    return this.itemRepository.findAll(spec, pageable);
  }

  protected Page<T> localFindAll(Pageable pageable) {
    return this.itemRepository.findAll(pageable);
  }

  protected T localSave(T item) {
    return this.itemRepository.save(item);
  }

  protected void localDelete(String id) {
    this.itemRepository.deleteById(id);
  }

  protected Page<T> localFindBySearchCriteria(T resource, String orOpredicateFlag,
          List<SearchCriteria> criterias,
          Pageable pageable) {
    if (criterias == null || criterias.isEmpty()) {
      return null;
    }
    Specification<T> spec = resource.getSearchSpecification(criterias.get(0));
    if (spec == null) {
      return null;
    }

    boolean isOrPredicate = orOpredicateFlag == null ? false : (orOpredicateFlag.equals(SearchOperation.OR_PREDICATE_FLAG));
    for (int i = 1; i < criterias.size(); i++) {
      if (isOrPredicate) {
        spec = Specification.where(spec)
                .or(resource.getSearchSpecification(criterias.get(i)));
      } else {
        spec = Specification.where(spec)
                .and(resource.getSearchSpecification(criterias.get(i)));
      }
    }
    return localFindAll(spec, pageable);

  }

  /************************************************
   * REST
   ************************************************/

  protected T remoteFindOne(String resId) {
    return this.itemRestService.findOne(resId);
  }

  protected Collection<T> remoteFindAll(Pageable pageable) {
    return this.itemRestService.findAll(pageable);
  }

  /**
   * Complete a page of resources by adding all its subresources
   * 
   * @param items
   */
  protected void completeEmbededResources(Page<T> items) {
    for (T item : items) {
      this.completeEmbededResources(item);
    }
  }

  /**
   * Complete a resource by adding all its subresources
   * 
   * @param item
   *          the resource to be completed
   */
  protected void completeEmbededResources(T item) {
    if (item instanceof SubResourceContainer) {
      List<Class<? extends Resource<?>>> embeddedResources = ((SubResourceContainer) item).getEmbeddedResourceTypes();
      for (Class<? extends Resource> type : embeddedResources) {
        this.completeEmbededResources(item, type);
      }
    }
  }

  /**
   * Complete a resource by adding a subresource of a given type
   * 
   * @param item
   *          the resource to be completed
   * @param subResourceType
   *          the subresource type
   * @param <V>
   *          The subresource type as a parameter type
   */
  protected <V extends Resource<V>> void completeEmbededResources(T item,
          Class<V> subResourceType) {

    if (item instanceof SubResourceContainer) {

      try {

        /*
         * Get sub resource Id from item property (i.e: getLanguageId)
         */
        String methodName = "get" + subResourceType.getSimpleName() + "Id";
        Method getter = item.getClass().getMethod(methodName);
        String subResourceId = (String) getter.invoke(item);

        if (!StringTool.isNullOrEmpty(subResourceId)) {

          Resource<?> child = this.getSubResource(subResourceId, subResourceType);

          if (child == null) {
            log.error(this.messageService.get("resources.embedded.notfound",
                    new Object[] { subResourceId, subResourceType.getName(), item.getResId(),
                            item.getClass().getName() }));
          }

          /*
           * Set the item property with fetched object
           */
          methodName = "set" + subResourceType.getSimpleName();
          Method setter = item.getClass().getMethod(methodName, subResourceType);
          setter.invoke(item, child);
        }

      } catch (Exception ex) {
        log.error("Unable to retrieve subresource " + subResourceType.getSimpleName() + " "
                + item.getResId());
      }
    } else {
      throw new IllegalArgumentException("item must be a SubResourceContainer");
    }
  }

  protected <S extends Resource<S>> ResourceService<S> getResourceService(Class<S> type) {
    throw new UnsupportedOperationException("Missing resource service for type " + type.getName());
  }

  /************************************************
   * Composit interface
   ************************************************/
  public Page<T> findBySearchCriteria(T resource, String orOpredicateFlag,
          List<SearchCriteria> criterias,
          Pageable pageable) {
    return this.localFindBySearchCriteria(resource, orOpredicateFlag, criterias, pageable);
  }

  public boolean existsById(String resId) {
    return this.localExistsById(resId);
  }

  public T findOne(String resId) {
    T item = this.localFindOne(resId);

    this.completeEmbededResources(item);

    return item;
  }

  public Page<T> findAll(Specification<T> spec, Pageable pageable) {

    Specification<T> completedSpec = this.addSpecificationFilter(spec);

    Page<T> items = this.localFindAll(completedSpec, pageable);

    this.completeEmbededResources(items);

    return items;
  }

  public Page<T> findAll(Pageable pageable) {
    Page<T> items = this.localFindAll(pageable);

    this.completeEmbededResources(items);

    return items;
  }

  /**
   * Find all subresources of a given type, with a specific parentId
   * 
   * @param parentId
   *          the id of the parent
   * @param parentClass
   *          the class of the resource
   * @param childClass
   *          the class of the subresource
   * @param pageable
   *          the pagination information
   * @param <V>
   *          the type of the subresource as a parameter type
   * @return
   */
  public <V> Page<V> findByParentId(String parentId, Class<?> parentClass, Class<V> childClass,
          Pageable pageable) {
    return this.itemRepository.findByParentId(parentId, parentClass, childClass, pageable);
  }

  public <V> V findByParentIdAndChildId(String parentId, Class<?> parentClass, Class<V> childClass,
          String id) {
    return this.itemRepository.findByParentIdAndChildId(parentId, parentClass, childClass, id);
  }

  public T save(T item) {
    item.init();
    this.beforeValidate(item);
    this.validate(item);

    return this.localSave(item);
  }

  public void delete(String id) {
    this.localDelete(id);
  }

  public <S extends Resource<S>> S getSubResource(String subResourceId, Class<S> subResourceType) {
    ResourceService<S> service = this.getResourceService(subResourceType);
    return service.findOneWithCache(subResourceId);
  }

  public <S extends Resource<S>> Collection<S> getSubResources(Iterable<String> subResourceIds,
          Class<S> subResourceType) {
    List<S> items = new ArrayList<>();

    if (subResourceIds != null) {
      for (String subResourceId : subResourceIds) {
        if (!StringTool.isNullOrEmpty(subResourceId)) {
          S item = this.getSubResource(subResourceId, subResourceType);
          if (item != null) {
            items.add(item);
          }
        }
      }
    }

    return new Collection<>(items);
  }

  /**************************************************/

  public SolidifyRepository<T> getRepository() {
    return this.itemRepository;
  }

  public ResourceService<T> getResourceService() {
    return this.itemRestService;
  }

  /**************************************************/

  /**
   * Override to add criteria to the Specification. May be useful to add criteria based on permissions
   * fetched from Authentication
   *
   * @param spec
   * @return
   */
  protected Specification<T> addSpecificationFilter(Specification<T> spec) {
    return spec;
  }

  /**
   * Override to set default data or clean properties
   *
   * @param item
   */
  protected void beforeValidate(T item) {
  }

  protected void validate(T item) {

    BindingResult errors = new BeanPropertyBindingResult(item, item.getClass().getName());

    this.validateConstraints(item, errors);
    this.linkedResourceValidator.validateLinkedResources(this, item, errors);
    this.validateItemSpecificRules(item, errors);

    if (errors.hasErrors()) {
      throw new SolidifyValidationException(new ValidationError(errors));
    }
  }

  /**
   * Validate Bean constraints
   *
   * @param RESTResource
   *          item
   * @param BindingResult
   *          errors
   */
  protected void validateConstraints(T item, BindingResult errors) {
    this.validator.validate(item, errors);
  }

  /**
   * Checks that all linked RESTResources identified by their resId exist. Must be overridden if the
   * Resource has linked RESTResources
   *
   * @param RESTResource
   *          item
   * @param BindingResult
   *          errors
   */
  public void validateLinkedResources(T item, BindingResult errors) {
  }

  /**
   * Override to add specific validation logic
   *
   * @param item
   * @param errors
   */
  protected void validateItemSpecificRules(T item, BindingResult errors) {
  }

}
