package ch.unige.solidify.service;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.domain.Sort.Order;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.UriComponentsBuilder;

import ch.unige.solidify.exception.RestResponseErrorHandler;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.exception.SolidifyRestException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.CollectionPage;
import ch.unige.solidify.rest.Result;
import ch.unige.solidify.util.CollectionTool;
import ch.unige.solidify.util.RestClientTool;

public abstract class SolidifyRestClientService {
  private static final Logger log = LoggerFactory.getLogger(SolidifyRestClientService.class);

  protected MessageService messageService;
  private RestClientTool restClientTool;
  private AuthenticationStorageProxy authenticationStorageProxy;

  public SolidifyRestClientService(RestClientTool restClientTool, MessageService messageService,
          AuthenticationStorageProxy authenticationStorageProxy) {
    this.restClientTool = restClientTool;
    this.messageService = messageService;
    this.authenticationStorageProxy = authenticationStorageProxy;
  }

  protected MultiValueMap<String, String> getAuthorizationHeaders() {
    if (this.authenticationStorageProxy.getAuthenticationStorage() != null) {
      return this.authenticationStorageProxy.getAuthenticationStorage()
              .getHeadersWithAuthorization();
    }
    return new LinkedMultiValueMap<>();
  }

  // *******************
  // ** GET Functions **
  // *******************

  protected String getResource(String url) {
    RestTemplate restClt = restClientTool.getClient(url);
    HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {
      ResponseEntity<String> responseEntity = restClt.exchange(url, HttpMethod.GET, httpEntity, String.class);
      return responseEntity.getBody();
    } catch (HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      String message = getResourceErrorMessage(url);
      log.error(message, e);
      throw new SolidifyRestException(message, e);
    }
  }

  protected <T> T getResource(String url, Class<T> clazz) {
    RestTemplate restClt = restClientTool.getClient(url);
    HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {
      ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.GET, httpEntity, clazz);
      return responseEntity.getBody();

    } catch (HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      String message = getResourceErrorMessage(url);
      log.error(message, e);
      throw new SolidifyRestException(message, e);
    }
  }

  protected String getResource(String url, Pageable pageable) {
    RestTemplate restClt = restClientTool.getClient(url);
    HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    final StringBuilder urlBuilder = new StringBuilder(url);
    // Add Pagination info
    if (pageable != null) {
      if (url.contains("?")) {
        urlBuilder.append("&");
      } else {
        urlBuilder.append("?");
      }
      urlBuilder.append(ActionName.PAGE + "=" + pageable.getPageNumber() + "&" + ActionName.SIZE + "=" + pageable.getPageSize());
      // Add Sort info
      if (pageable.getSort() != null) {
        pageable.getSort().stream()
                .forEach(
                        s -> urlBuilder.append("&" + ActionName.SORT + "=" + s.getProperty() + "," + s.getDirection().toString().toLowerCase()));
      }
    }
    url = urlBuilder.toString();
    try {
      logModuleCall(url);
      ResponseEntity<String> responseEntity = restClt.exchange(url, HttpMethod.GET, httpEntity, String.class);
      return responseEntity.getBody();
    } catch (HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      String message = getResourceErrorMessage(url);
      log.error(message, e);
      throw new SolidifyRestException(message, e);
    }
  }

  protected String getResources(String url, Pageable pageable) {
    url = this.completeUrlWithPagination(url, pageable);

    RestTemplate restClt = restClientTool.getClient(url);
    HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {

      logModuleCall(url);

      return restClt.exchange(url, HttpMethod.GET, httpEntity, String.class).getBody();
    } catch (HttpClientErrorException e) {

      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      } else {
        String message = getResourceErrorMessage(url);
        log.error(message, e);
        throw new SolidifyRestException(message, e);
      }
    }
  }

  /**
   * @param <T>
   * @param url
   * @param clazz
   * @param pageable
   * @return
   */
  public <T> Collection<T> getResourceList(String url, Class<T> clazz, Pageable pageable) {
    String jsonResult = this.getResources(url, pageable);
    List<T> list = CollectionTool.getList(jsonResult, clazz);
    CollectionPage page = CollectionTool.getPage(jsonResult);
    return new Collection<>(list, page);
  }

  // ********************
  // ** HEAD Functions **
  // ********************

  protected HttpStatus headResource(String url) {
    RestTemplate restClt = restClientTool.getClient(url);
    HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {
      restClt.exchange(url, HttpMethod.HEAD, httpEntity, Void.class);
    } catch (HttpClientErrorException e) {
      return e.getStatusCode();
    }
    return HttpStatus.OK;
  }

  // ********************
  // ** POST Functions **
  // ********************

  protected <T> T postResource(String url, Class<T> clazz) {
    RestTemplate restClt = restClientTool.getClient(url);
    HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {

      ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.POST, httpEntity, clazz);
      return responseEntity.getBody();

    } catch (HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      String message = getResourceErrorMessage(url);
      log.error(message, e);
      throw new SolidifyRestException(message, e);
    }
  }

  protected Result postResourceAction(String url) {
    return postResource(url, Result.class);
  }

  protected <T> T postResource(String url, T newResource, Class<T> clazz) {
    RestTemplate restClt = restClientTool.getClient(url);
    restClt.setErrorHandler(new RestResponseErrorHandler());
    HttpEntity<?> httpEntity = new HttpEntity<>(newResource, this.getAuthorizationHeaders());
    try {
      ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.POST, httpEntity, clazz);
      return responseEntity.getBody();
    } catch (HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      String message = getResourceErrorMessage(url);
      log.error(message, e);
      throw new SolidifyRestException(message, e);
    }
  }

  protected <T, V> T postResourceOther(String url, V otherResource, Class<T> clazz) {
    RestTemplate restClt = restClientTool.getClient(url);
    restClt.setErrorHandler(new RestResponseErrorHandler());
    HttpEntity<?> httpEntity = new HttpEntity<>(otherResource, this.getAuthorizationHeaders());
    try {
      ResponseEntity<T> responseEntity = restClt.exchange(url, HttpMethod.POST, httpEntity, clazz);
      return responseEntity.getBody();
    } catch (HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      String message = getResourceErrorMessage(url);
      log.error(message, e);
      throw new SolidifyRestException(message, e);
    }
  }

  // **********************
  // ** DELETE Functions **
  // **********************

  protected void deleteResource(String url) {
    RestTemplate restClt = restClientTool.getClient(url);
    restClt.setErrorHandler(new RestResponseErrorHandler());
    HttpEntity<?> httpEntity = new HttpEntity<>(this.getAuthorizationHeaders());
    try {
      restClt.exchange(url, HttpMethod.DELETE, httpEntity, Void.class);
    } catch (HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      String message = getResourceErrorMessage(url);
      log.error(message, e);
      throw new SolidifyRestException(message, e);
    }
  }

  // *******************
  // ** PUT Functions **
  // *******************

  protected <T> T putResource(String url, T updatedResource, Class<T> clazz) {
    RestTemplate restClt = restClientTool.getClient(url);
    restClt.setErrorHandler(new RestResponseErrorHandler());
    HttpEntity<?> httpEntity = new HttpEntity<>(updatedResource, this.getAuthorizationHeaders());
    try {
      return restClt.exchange(url, HttpMethod.PUT, httpEntity, clazz).getBody();
    } catch (HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new SolidifyResourceNotFoundException(url);
      }
      String message = getResourceErrorMessage(url);
      log.error(message, e);
      throw new SolidifyRestException(message, e);
    }
  }

  // *********************
  // ** Misc. Functions **
  // *********************

  protected String completeUrlWithPagination(String url, Pageable pageable) {
    UriComponentsBuilder uriBuilder = UriComponentsBuilder.fromUriString(url);

    if (pageable != null) {

      int page = pageable.getPageNumber();
      int size = pageable.getPageSize();
      Sort sort = pageable.getSort();

      MultiValueMap<String, String> params = new LinkedMultiValueMap<>();
      params.set("page", Integer.toString(page));
      params.set("size", Integer.toString(size));

      if (sort != null) {
        // TODO: support sorting on many fields ?
        for (Order order : sort) {
          String sortProperty = order.getProperty();
          String sortDirection = order.getDirection().toString();
          params.set("sort", sortProperty + "," + sortDirection);
        }
      }

      uriBuilder.queryParams(params);
    }

    return uriBuilder.toUriString();
  }

  private String getResourceErrorMessage(String url) {
    return "Error in getting resource (" + url + ")";
  }

  private void logModuleCall(String moduleUrl) {
    try {
      URL url = new URL(moduleUrl);
      String userInfo = url.getUserInfo();
      if (userInfo != null) {
        String userInfoWithoutPassword = userInfo.substring(0, userInfo.indexOf(':')) + ":**********";
        String urlWithoutPassword = url.getProtocol() + "://" + userInfoWithoutPassword + "@" + url.getHost() + url
                .getPort() + url.getPath() + "?" + url.getQuery();
        log.info("Module call GET: {}", urlWithoutPassword);
      } else {
        log.info("Module call GET: {}", moduleUrl);
      }
    } catch (MalformedURLException e) {
      log.error("Malformed url : {}", moduleUrl, e);
    }
  }

}
