package ch.unige.solidify.test.doc;

import static ch.unige.solidify.test.executor.SolidifyExecutorTestDoc.makeFieldsDocumentation;
import static ch.unige.solidify.test.executor.SolidifyExecutorTestDoc.makeLinksDocumentation;
import static ch.unige.solidify.test.executor.SolidifyExecutorTestDoc.performExpectationTests;
import static ch.unige.solidify.test.executor.SolidifyExecutorTestDoc.performRequest;
import static ch.unige.solidify.test.helper.RestDocsTemplates.HREF_FIELD;
import static ch.unige.solidify.test.helper.RestDocsTemplates.LINKS_FIELD;
import static ch.unige.solidify.test.helper.RestDocsType.APPLICATION;
import static ch.unige.solidify.test.helper.RestDocsType.MODULE;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.fail;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.documentationConfiguration;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Predicate;

import org.json.JSONObject;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.MediaType;
import org.springframework.restdocs.JUnitRestDocumentation;
import org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.context.WebApplicationContext;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.controller.ModuleController;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.test.helper.AbstractResourceDoc;
import ch.unige.solidify.test.helper.RestDocsTemplates;
import ch.unige.solidify.test.helper.RestDocsType;
import ch.unige.solidify.test.model.OpenApiExtraDocumentation;

/**
 * Generic Rest-Docs context configuration for Solidify applications
 */
public abstract class RestDocsDocumentation {

  private static final String URL_SEPARATOR = "/";
  private final Logger log = LoggerFactory.getLogger(RestDocsDocumentation.class);
  private Map<URI, AbstractResourceDoc> resources = new HashMap<>();
  private MockMvc mockMvc;
  private Map<String, String> ids = new HashMap<>();

  protected List<String> urlIgnoreList = new ArrayList<>();

  protected List<String> urlReadOnlyList = new ArrayList<>();

  protected Gson gson;

  @Rule
  public final JUnitRestDocumentation restDocumentation = new JUnitRestDocumentation();

  @Autowired
  private WebApplicationContext webApplicationContext;

  @Autowired
  private List<ModuleController> moduleControllerList;

  @Autowired
  RestDocsTemplates restDocsTemplates;

  @Value("${server.servlet.context-path}")
  private String contextPath;

  private List<OpenApiExtraDocumentation> listOpenApiExtraDocumentation;

  /**
   * Used to bypass no testable url's.
   */
  protected abstract void setUrlIgnoreList();

  protected abstract void setUrlReadOnlyList();

  @Before
  public void setUp() {
    this.gson = new GsonBuilder().create();
    this.mockMvc = MockMvcBuilders.webAppContextSetup(this.webApplicationContext).defaultRequest(getApplicationRequestBuilder())
            .apply(documentationConfiguration(this.restDocumentation)).build();
    setUrlIgnoreList();
    log.info("Total ignored URLs {}", this.urlIgnoreList.size());
    setUrlReadOnlyList();
    log.info("Total Read-Only URLs {}", this.urlReadOnlyList.size());
    loadResources();
  }

  protected abstract void loadResources();

  protected void addResources(AbstractResourceDoc r) {
    this.resources.put(URI.create(contextPath + r.getName()), r);
  }

  /**
   * Return a request builder for applications
   */
  private MockHttpServletRequestBuilder getApplicationRequestBuilder() {
    return RestDocumentationRequestBuilders.get(contextPath).contextPath(contextPath);
  }

  /**
   * Return a request builder for modules
   */
  private MockHttpServletRequestBuilder getModuleRequestBuilder(String module) {
    String url = getModuleUrl(module);
    log.info("Building module level request [{}]", url);
    MockHttpServletRequestBuilder mockHttpServletRequestBuilder = RestDocumentationRequestBuilders.get(url);
    mockHttpServletRequestBuilder.contextPath(contextPath);
    return mockHttpServletRequestBuilder;
  }

  /**
   * Return a request builder for resourcesTestObjectBuilder
   */
  private MockHttpServletRequestBuilder getResourceRequestBuilder(String resource, RequestMethod method, JSONObject content) {
    log.info("Building resource level request [{}], method=[{}]", resource, method);
    MockHttpServletRequestBuilder mockHttpServletRequestBuilder;
    if (method.equals(RequestMethod.POST) && content != null) {
      mockHttpServletRequestBuilder = RestDocumentationRequestBuilders
              .post(resource).content(content.toString()).contentType(MediaType.APPLICATION_JSON);
    } else {
      assertNull(content);
      mockHttpServletRequestBuilder = RestDocumentationRequestBuilders.get(resource);
    }
    return mockHttpServletRequestBuilder.contextPath(contextPath);
  }

  /**
   * Get the modules list corresponding with the current application
   */
  private List<String> getModules() {
    List<String> modules = new ArrayList<>();
    moduleControllerList.forEach(moduleController -> {
      modules.add(moduleController.getRoot());
    });
    return modules;
  }

  /**
   * @return the context path
   */
  protected String getContextPath() {
    return contextPath;
  }

  /**
   * Fetch resources from json response
   */
  private List<String> fetchResources(ResultActions resultActions, String resourceType) throws IOException {
    JsonNode json = new ObjectMapper().readTree(resultActions.andReturn().getResponse().getContentAsString());
    if (json.path(LINKS_FIELD).get(resourceType) == null) {
      return new ArrayList<>();
    }
    return json.path(LINKS_FIELD).get(resourceType).findValuesAsText(HREF_FIELD);
  }

  /**
   * Build and execute a mock request
   */
  private ResultActions executeMockRequest(RestDocsType restDocsType, String path, RequestMethod method) {
    return executeMockRequest(restDocsType, path, method, null);
  }

  private ResultActions executeMockRequest(RestDocsType restDocsType, String path, RequestMethod method, AbstractResourceDoc resDoc) {
    MockHttpServletRequestBuilder mockHttpServletRequestBuilder = null;
    ResultActions result = null;
    int expectedSize = 0;
    switch (restDocsType) {
      case RESOURCE_LIST:
        mockHttpServletRequestBuilder = getResourceRequestBuilder(path, RequestMethod.GET, null);
        expectedSize = restDocsTemplates.getCommonResourceListFields().size();
        break;
      case RESOURCE:
        if (resDoc == null) {
          throw new SolidifyRuntimeException("Missing resource documentation");
        }
        if (RequestMethod.POST.equals(method)) {
          mockHttpServletRequestBuilder = getResourceRequestBuilder(path, RequestMethod.POST, resDoc.getJsonFields(ids));
        } else {
          mockHttpServletRequestBuilder = getResourceRequestBuilder(path, RequestMethod.GET, null);
        }
        expectedSize += resDoc.getNumberOfExpectedReadOnlyFields() + restDocsTemplates.getCommonResourceFields().size();
        break;
      case MODULE:
        mockHttpServletRequestBuilder = getModuleRequestBuilder(path);
        expectedSize = restDocsTemplates.getCommonModuleFields().size();
        break;
      case APPLICATION:
        mockHttpServletRequestBuilder = getApplicationRequestBuilder();
        expectedSize = restDocsTemplates.getCommonApplicationFields().size();
        break;
    }
    try {
      result = performRequest(mockMvc, mockHttpServletRequestBuilder, method);
      JsonNode jsonMockResponse = new ObjectMapper().readTree(result.andReturn().getResponse().getContentAsString());
      // Id created
      if (restDocsType == RestDocsType.RESOURCE) {
        ids.put(resDoc.getClass().getSimpleName(), jsonMockResponse.get("resId").textValue());
      }
      assertEquals(path, expectedSize, jsonMockResponse.size());
    } catch (IOException e) {
      fail("Failure getting json response content. [" + result.andReturn().getResponse() + "]");
    }
    return result;
  }

  private boolean isIgnoredUrl(String url) {
    Predicate<String> urlIgnoredPredicate = predicate -> this.urlIgnoreList.contains(url);
    Optional<String> optional = urlIgnoreList.stream().filter(urlIgnoredPredicate).findFirst();
    if (optional.isPresent()) {
      log.warn("Ignored URL [{}]", url);
      return true;
    }
    return false;
  }

  private boolean isReadOnlyUrl(String url) {
    Predicate<String> urlReadOnlyPredicate = predicate -> this.urlReadOnlyList.contains(url);
    Optional<String> optional = urlIgnoreList.stream().filter(urlReadOnlyPredicate).findFirst();
    if (optional.isPresent()) {
      log.warn("Read-Only URL [{}]", url);
      return true;
    }
    return false;
  }

  /**
   * The main REST calls of each application
   */
  @Test
  public void generateMainRESTDocs() {
    ResultActions result = executeMockRequest(APPLICATION, null, RequestMethod.GET);
    performExpectationTests(result, restDocsTemplates.getApplicationTestTemplate(getModules()));
    String uri = result.andReturn().getRequest().getRequestURI();
    makeLinksDocumentation(result, uri, false, restDocsTemplates.getApplicationLinkDescriptors(getModules()));
    makeFieldsDocumentation(result, uri, restDocsTemplates.getApplicationFieldsTemplate(), true);
  }

  /**
   * The main REST calls of each module
   */
  @Test
  public void generateModulesRESTDocs() throws IOException {
    for (String module : getModules()) {
      if (isIgnoredUrl(getModuleUrl(module))) {
        continue;
      }
      ResultActions result = executeMockRequest(MODULE, module, RequestMethod.GET);
      String uri = result.andReturn().getRequest().getRequestURI();
      List<String> resourcesFromResult = fetchResources(result, RestDocsTemplates.RESOURCES_FIELD);
      List<String> toolsFromResult = fetchResources(result, RestDocsTemplates.TOOLS_FIELD);
      if (!resourcesFromResult.isEmpty()) {
        makeLinksDocumentation(result, uri, false, restDocsTemplates.getModulesLinkDescriptors(module, resourcesFromResult, toolsFromResult));
        makeFieldsDocumentation(result, uri, restDocsTemplates.getModuleFieldsDescriptor(module), true);
      }
    }
  }

  /**
   * The main REST calls of each resource
   */
  @Test
  public void generateResourcesRESTDocs() throws IOException {
    listOpenApiExtraDocumentation = new ArrayList<>();
    for (String module : getModules()) {
      if (isIgnoredUrl(getModuleUrl(module))) {
        continue;
      }
      ResultActions resultActions = executeMockRequest(MODULE, module, RequestMethod.GET);
      // Data resource generation
      fetchResources(resultActions, RestDocsTemplates.RESOURCES_FIELD).stream().forEach(resource -> {
        URI uri = URI.create(URI.create(resource).getPath());
        String resourcePath = uri.getPath();
        ResultActions resourceResult = null;
        ResultActions listResult;
        if (!isIgnoredUrl(resourcePath)) {
          if (this.resources.get(uri) != null) {
            AbstractResourceDoc docTemplate = this.resources.get(uri);
            if (!isReadOnlyUrl(resourcePath)) {
              // Create a resource
              resourceResult = executeMockRequest(RestDocsType.RESOURCE, resource, RequestMethod.POST, docTemplate);
            } else {
              // Find first resource
              try {
                String resId = this.getFirstResId(
                        executeMockRequest(RestDocsType.RESOURCE_LIST, resource, RequestMethod.GET, docTemplate).andReturn().getResponse()
                                .getContentAsString());
                resourceResult = executeMockRequest(RestDocsType.RESOURCE, resource + SolidifyConstants.URL_SEP + resId, RequestMethod.GET,
                        docTemplate);
              } catch (UnsupportedEncodingException e) {
                throw new SolidifyRuntimeException(e.getMessage());
              }
            }
            // List resource collection
            listResult = executeMockRequest(RestDocsType.RESOURCE_LIST, resource, RequestMethod.GET);

            // Document resource fields
            makeFieldsDocumentation(resourceResult, resourcePath,
                    restDocsTemplates.getResourceFieldDescriptors(resourcePath, docTemplate.getReadOnlyDocumentationFields()),
                    isReadOnlyUrl(resourcePath));
            // Document resource links
            makeLinksDocumentation(resourceResult, resourcePath, false,
                    restDocsTemplates.getResourceDataLinkDescriptors(resourcePath, docTemplate.getResourceLinks()));
            // Document resource collection links
            makeLinksDocumentation(listResult, resourcePath, true,
                    restDocsTemplates.getResourceLinkDescriptors(module, resourcePath, docTemplate.getListLinks()));
            registerExtraDocumentation(docTemplate.getOpenApiExtraDocumentation());
          } else {
            // Resource to document
            log.warn("No documentation for {}", resourcePath);
            // List resource collection
            listResult = executeMockRequest(RestDocsType.RESOURCE_LIST, resource, RequestMethod.GET);
            makeFieldsDocumentation(listResult, resourcePath, restDocsTemplates.getResourceCollectionFieldDescriptors(),
                    isReadOnlyUrl(resourcePath));
          }
        }
      });
    }
    generateOpenApiExtraDocumentation();
  }

  private String getFirstResId(String contentAsString) {
    ObjectMapper mapper = new ObjectMapper();
    mapper.registerModule(new com.fasterxml.jackson.datatype.jsr310.JavaTimeModule());
    mapper.configure(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS, false);
    mapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

    try {
      JsonNode jsonData = mapper.readTree(contentAsString).get("_data");
      if (jsonData.isArray()) {
        for (JsonNode json : jsonData) {
          if (json.get("resId") != null) {
            return json.get("resId").textValue();
          }
        }
      }

    } catch (IOException e) {
      throw new SolidifyRuntimeException(e.getMessage());
    }
    throw new SolidifyRuntimeException("Cannot find first resId");
  }

  private void registerExtraDocumentation(OpenApiExtraDocumentation openApiExtraDocumentation) {
    listOpenApiExtraDocumentation.add(openApiExtraDocumentation);
  }

  private void generateOpenApiExtraDocumentation() {
    String jsonExtraDocumentation = this.gson.toJson(listOpenApiExtraDocumentation);
    try {
      saveJsonFile("target/restdocs-spec/openapi-extra-documentation.json", jsonExtraDocumentation);
    } catch (IOException e) {
      log.error("IO exception", e);
      fail("IO exception: " + e.getMessage());
    }
  }

  public void saveJsonFile(String outputFilePath, String json) throws IOException {
    createDirectory(outputFilePath);
    Path jsonFile = Paths.get(outputFilePath);
    // Use try-with-resource to get auto-closeable writer instance
    try (BufferedWriter writer = Files.newBufferedWriter(jsonFile)) {
      writer.write(json);
    }
    System.out.println("Successfully Copied JSON Object into : " + outputFilePath);
  }

  private void createDirectory(String outputFilePath) {
    File directory = new File(getDirectoryPath(outputFilePath));
    if (!directory.exists()) {
      directory.mkdirs();
    }
  }

  private String getDirectoryPath(String outputFilePath) {
    if (outputFilePath.endsWith(URL_SEPARATOR))
      return outputFilePath;
    int lastIndexOfSeparator = outputFilePath.lastIndexOf(URL_SEPARATOR);
    return outputFilePath.substring(0, lastIndexOfSeparator);
  }

  private String getModuleUrl(String module) {
    return contextPath + SolidifyConstants.URL_SEP + module;
  }

}
