package ch.unige.solidify.specification;

import static ch.unige.solidify.util.SearchOperation.BETWEEN;

import java.math.BigDecimal;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.OffsetDateTime;

import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Expression;
import javax.persistence.criteria.Path;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.util.StringUtils;

import ch.unige.solidify.util.SearchCriteria;

/**
 * Specification generated with a SearchCriteria.
 *
 * @param <T>
 * @see ch.unige.solidify.util.SearchCriteria
 */
public abstract class SearchSpecification<T> implements Specification<T> {
  private static final long serialVersionUID = 7428274476634034865L;

  private static final Logger log = LoggerFactory.getLogger(SearchSpecification.class);

  private SearchCriteria criteria;

  protected SearchSpecification(SearchCriteria criteria) {
    this.criteria = criteria;
  }

  @Override

  /**
   * Creates a WHERE clause for a query of the referenced entity in form of a Predicate for the given
   * Root and instance variable SearchCriteria "criteria". criteria.key is one of entity field name
   * criteria.value is used to create instance of different Object according to the entity field type
   * criteria.operationType is used to choose predicate
   * 
   * @see ch.unige.solidify.util.SearchCriteria
   */
  public Predicate toPredicate(Root<T> root, CriteriaQuery<?> query, CriteriaBuilder builder) {
    Path fieldPath = findFieldPath(root, criteria.getKey());
    Class fieldType = fieldPath.getJavaType();
    Object criteriaValue = null;
    if (!criteria.getOperationType().equals(BETWEEN)) {
      criteriaValue = convertString2Object(fieldType, criteria.getValue());
    }
    switch (criteria.getOperationType()) {
      case EQUALITY:
        if (criteria.isCaseInsensitive() && isString(fieldType)) {
          return builder.equal(
                  builder.lower(fieldPath),
                  criteria.getValue().toString().toLowerCase());
        }
        return builder.equal(fieldPath, criteriaValue);

      case NEGATION:
        if (criteria.isCaseInsensitive() && isString(fieldType)) {
          return builder.notEqual(
                  builder.lower(fieldPath),
                  criteria.getValue().toString().toLowerCase());
        }
        return builder.notEqual(fieldPath, criteriaValue);

      case GREATER_THAN:
        if (criteria.isCaseInsensitive() && isString(fieldType)) {
          return builder.greaterThan(
                  builder.lower(fieldPath),
                  criteria.getValue().toString().toLowerCase());
        }
        return builder.greaterThan(fieldPath, (Comparable) criteriaValue);

      case LESS_THAN:
        if (criteria.isCaseInsensitive() && isString(fieldType)) {
          return builder.lessThan(
                  builder.lower(fieldPath),
                  criteria.getValue().toString().toLowerCase());
        }
        return builder.lessThan(fieldPath, (Comparable) criteriaValue);

      case STARTS_WITH:
        if (!isString(fieldType))
          return null;
        if (criteria.isCaseInsensitive() && isString(fieldType)) {
          return builder.like(
                  builder.lower(fieldPath),
                  criteria.getValue().toString().toLowerCase() + "%");
        }
        return builder.like((Expression<String>) fieldPath, criteria.getValue() + "%");

      case ENDS_WITH:
        if (criteria.isCaseInsensitive() && isString(fieldType)) {
          return builder.like(
                  builder.lower(fieldPath),
                  "%" + criteria.getValue().toString().toLowerCase());
        }
        return builder.like((Expression<String>) fieldPath, "%" + criteria.getValue());

      case CONTAINS:
        if (criteria.isCaseInsensitive() && isString(fieldType)) {
          return builder.like(
                  builder.lower(fieldPath),
                  "%" + criteria.getValue().toString().toLowerCase() + "%");
        }
        return builder.like((Expression<String>) fieldPath, "%" + criteria.getValue() + "%");

      case ORACLE_CONTAINS:
        Expression<Integer> exp = builder.function("CONTAINS", Integer.class, fieldPath,
                builder.literal(criteria.getValue().toString()),
                builder.literal(1));
        return (builder.gt(exp, 0));

      case GREATER_EQUAL:
        if (criteria.isCaseInsensitive() && isString(fieldType)) {
          return builder.greaterThanOrEqualTo(
                  builder.lower(fieldPath),
                  criteria.getValue().toString().toLowerCase());
        }
        return builder.greaterThanOrEqualTo(fieldPath, (Comparable) criteriaValue);

      case LESS_EQUAL:
        if (criteria.isCaseInsensitive() && isString(fieldType)) {
          return builder.lessThanOrEqualTo(
                  builder.lower(fieldPath),
                  criteria.getValue().toString().toLowerCase());
        }
        return builder.lessThanOrEqualTo(fieldPath, (Comparable) criteriaValue);

      case BETWEEN:
        if (isString(fieldType))
          throw new RuntimeException("BETWEEN operation can be applied only for number or date");
        Object[] values = convertBetweenValueString2Object(fieldType, criteria.getValue());
        return builder
                .between(fieldPath, (Comparable) values[0], (Comparable) values[1]);
      default:
        return null;
    }
  }

  private Path<?> findFieldPath(Root<?> root, String fieldName) {
    Path<?> expression;
    if (fieldName.contains(".")) {
      String[] names = StringUtils.split(fieldName, ".");
      expression = root.get(names[0]);
      for (int i = 1; i < names.length; i++) {
        expression = expression.get(names[i]);
      }
    } else {
      expression = root.get(fieldName);
    }
    return expression;
  }

  private boolean isString(Class fieldType) {
    return String.class.isAssignableFrom(fieldType);
  }

  // for between operator, the value will be 20_30 or 2016-08-16_2018-09-25.
  // it treats only non-string field
  private Object[] convertBetweenValueString2Object(Class fieldType, Object inputValue) {
    String[] values = StringUtils.split((String) inputValue, "_");
    Object first = convertString2Object(fieldType, values[0]);
    Object second = convertString2Object(fieldType, values[1]);
    Object[] o = { first, second };
    return o;
  }

  private Object convertString2Object(Class fieldType, Object inputValue) {
    // first test number
    if (Number.class.isAssignableFrom(fieldType)) {
      return new BigDecimal((String) inputValue);
    }

    // Test boolean
    if (Boolean.class.isAssignableFrom(fieldType) || boolean.class.isAssignableFrom(fieldType)) {
      return Boolean.parseBoolean((String) inputValue);
    }

    // Obtains an instance of LocalDate from a text string such as "2016-08-16"
    // The string must represent a valid date-time and is parsed using DateTimeFormatter.ISO_LOCAL_DATE.
    if (LocalDate.class.isAssignableFrom(fieldType)) {
      return LocalDate.parse((String) inputValue);
    }
    // Obtains an instance of LocalDateTime from a text string such as 2007-12-03T10:15:30.
    // The string must represent a valid date-time and is parsed using
    // DateTimeFormatter.ISO_LOCAL_DATE_TIME.
    if (LocalDateTime.class.isAssignableFrom(fieldType)) {
      return LocalDateTime.parse((String) inputValue);
    }
    // Obtains an instance of OffsetDateTime from a text string such as 2007-12-03T10:15:30+01:00
    if (OffsetDateTime.class.isAssignableFrom(fieldType)) {
      return OffsetDateTime.parse((String) inputValue);
    }
    return inputValue;
  }

}
