package ch.unige.solidify.util;

import java.io.IOException;
import java.net.URI;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.boot.web.client.RestTemplateCustomizer;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.http.client.ClientHttpRequestInterceptor;
import org.springframework.http.client.ClientHttpResponse;
import org.springframework.http.client.SimpleClientHttpRequestFactory;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.StringHttpMessageConverter;
import org.springframework.stereotype.Component;
import org.springframework.web.client.DefaultResponseErrorHandler;
import org.springframework.web.client.RequestCallback;
import org.springframework.web.client.ResponseExtractor;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.interceptor.BearerAuthorizationInterceptor;
import ch.unige.solidify.interceptor.HeaderRequestInterceptor;
import ch.unige.solidify.interceptor.PlusEncoderInterceptor;
import ch.unige.solidify.model.sso.SsoUser;
import ch.unige.solidify.security.TrustedTokenStorage;

@Component
public class RestClientTool {
  private static final Logger log = LoggerFactory.getLogger(RestClientTool.class);

  @Value("${solidify.oauth2.accesstoken:#{null}}")
  private String accessToken;

  private final TrustedTokenStorage trustedTokenStorage;

  public RestClientTool(TrustedTokenStorage trustedTokenStorage) {
    this.trustedTokenStorage = trustedTokenStorage;
    if (accessToken == null) {
      accessToken = System.getProperty("solidify.oauth2.accesstoken");
    }
  }

  public RestTemplate getClient(String url) {
    return getRestTemplateBuilder(url).build();
  }

  public RestTemplate getClientWithPlusEncoderInterceptor(String url) {
    return getRestTemplateBuilder(url).additionalInterceptors(new PlusEncoderInterceptor()).build();
  }

  public <T> T downloadContent(URI uri, ResponseExtractor<T> responseExtractor) {
    RestTemplate restTemplate = getRestTemplateBuilder(uri.toString()).requestFactory(() -> {
      SimpleClientHttpRequestFactory requestFactory = new SimpleClientHttpRequestFactory();
      requestFactory.setBufferRequestBody(false);
      return requestFactory;
    }).build();

    // Optional Accept header
    RequestCallback requestCallback = request -> request.getHeaders()
            .setAccept(Arrays.asList(MediaType.APPLICATION_OCTET_STREAM, MediaType.ALL));

    return restTemplate.execute(uri, HttpMethod.GET, requestCallback, responseExtractor);
  }

  private RestTemplateBuilder getRestTemplateBuilder(String urlString) {
    RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
    if (!StringTool.isNullOrEmpty(accessToken)) {
      // OAuth2 without end user: shared AccesToken
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString)
              .additionalInterceptors(new BearerAuthorizationInterceptor(accessToken));
    } else if (trustedTokenStorage.getTrustedToken() != null) {
      // DLCM : Access token of trusted user
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString).additionalInterceptors(
              new BearerAuthorizationInterceptor(trustedTokenStorage.getTrustedToken()));
    } else if (UserTool.getUserPrincipal() instanceof SsoUser) {
      // Rock Solid : AccesToken, usercontext, useruniqueID of end user for OIDC Autho2 protected MS
      SsoUser user = (SsoUser) UserTool.getUserPrincipal();
      List<ClientHttpRequestInterceptor> intercepters = new ArrayList<>();
      if (user.getUniqueId() != null) {
        intercepters.add(new HeaderRequestInterceptor(SsoUser.CONNECTED_USER_UNIQUE_ID,
                user.getUniqueId()));
      }
      if (user.getUserContext() != null) {
        intercepters.add(new HeaderRequestInterceptor(SsoUser.CONNECTED_USER_CONTEXT_SOURCE_NAME,
                user.getUserContext()));
      }
      if (UserTool.getUserPrincipal().getAccessToken() != null) {
        intercepters.add(new BearerAuthorizationInterceptor(
                UserTool.getUserPrincipal().getAccessToken()));
      }
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString)
              .additionalInterceptors(intercepters);
    } else {
      // No Auth
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString);
    }

    // Use UTF-8 message converter
    restTemplateBuilder = restTemplateBuilder.additionalCustomizers(new UTF8MessageConverterCustomizer());

    // Error handler
    restTemplateBuilder.errorHandler(new DefaultResponseErrorHandler() {
      @Override
      public void handleError(ClientHttpResponse response) throws IOException {
        log.error("Error {} on {}", response.getStatusText(), getUrl(new URL(urlString)));
        super.handleError(response);
      }
    });
    return restTemplateBuilder;
  }

  private String getUrl(URL url) {
    String rootUrl = url.getProtocol() + "://" + url.getHost();
    if (url.getPort() > 0) {
      rootUrl += ":" + url.getPort();
    }
    return rootUrl + url.getPath();
  }

  private static class UTF8MessageConverterCustomizer implements RestTemplateCustomizer {
    public void customize(RestTemplate restTemplate) {
      List<HttpMessageConverter<?>> converters = restTemplate.getMessageConverters();
      converters.removeIf(httpMessageConverter -> httpMessageConverter instanceof StringHttpMessageConverter);
      converters.add(0, new StringHttpMessageConverter(StandardCharsets.UTF_8));
    }
  }

}
