package ch.unige.solidify.util;

import java.io.Serializable;

/**
 * This is class is used to present a search criteria. It contains the key (normally it is an entity
 * filed name), value (expected value) and operator (less than , equals...) Optionally it contains a
 * flag who indicate if the value is case sensitive or not. The default behavior is case sensitive.
 */
public class SearchCriteria implements Serializable {
  private static final long serialVersionUID = -6340839586653964229L;
  private String key;
  private SearchOperation operationType;
  private String operation;
  private Object value;
  private String caseType;
  private boolean caseInsensitive = false;

  public SearchCriteria() {
    super();
  }

  /**
   * @param key
   *          normally it is an entity filed name
   * @param operation
   *          operation type string value @see ch.unige.solidify.util.SearchOperation
   * @param value
   *          expected value
   */
  public SearchCriteria(final String key, final String operation, final Object value) {
    super();
    this.operation = operation;
    this.key = key;
    this.operationType = SearchOperation.getOperation(operation);
    this.value = value;
  }

  /**
   * @param key
   *          normally it is an entity filed name
   * @param operationType
   *          operation type @see ch.unige.solidify.util.SearchOperation
   * @param value
   *          expected value
   */
  public SearchCriteria(final String key, SearchOperation operationType, final Object value) {
    super();
    this.key = key;
    this.operationType = operationType;
    this.value = value;
  }

  /**
   * @param caseType
   *          "i" or "c" i means case Insensitive
   * @param key
   *          normally it is an entity filed name
   * @param operation
   *          operation type string value @see ch.unige.solidify.util.SearchOperation
   * @param value
   *          expected value
   */
  public SearchCriteria(final String caseType, final String key, final String operation,
          final Object value) {
    super();
    this.caseInsensitive = (caseType != null
            && caseType.equals(SearchOperation.CASE_INSENSITIVE_FLAG));
    this.key = key;
    this.operation = operation;
    this.operationType = SearchOperation.getOperation(operation);
    this.value = value;
  }

  /**
   * @param caseType
   *          "i" or "c" i means case Insensitive
   * @param key
   *          normally it is an entity filed name
   * @param operationType
   *          operation type @see ch.unige.solidify.util.SearchOperation
   * @param value
   *          expected value
   */

  public SearchCriteria(final String caseType, final String key,
          final SearchOperation operationType, final Object value) {
    super();
    this.caseInsensitive = (caseType != null
            && caseType.equals(SearchOperation.CASE_INSENSITIVE_FLAG));
    this.key = key;
    this.operationType = operationType;
    this.value = value;
  }

  public SearchOperation getOperationType() {
    return operationType;
  }

  public String getOperation() {
    return operation;
  }

  public void setOperation(String operation) {
    this.operation = operation;
    this.operationType = SearchOperation.getOperation(operation);
  }

  public String getKey() {
    return key;
  }

  public void setKey(String key) {
    this.key = key;
  }

  public Object getValue() {
    return value;
  }

  public void setValue(Object value) {
    this.value = value;
  }

  public boolean isCaseInsensitive() {
    return caseInsensitive;
  }

  public void setCaseInsensitive(boolean caseInsensitive) {
    this.caseInsensitive = caseInsensitive;
  }

  public String getCaseType() {
    return caseType;
  }

  public void setCaseType(String caseType) {
    this.caseType = caseType;
    this.caseInsensitive = (caseType != null
            && caseType.equals(SearchOperation.CASE_INSENSITIVE_FLAG));
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (getOperationType() != null ? getOperationType().hashCode() : 0);
    result = 31 * result + (getOperation() != null ? getOperation().hashCode() : 0);
    result = 31 * result + (getKey() != null ? getKey().hashCode() : 0);
    result = 31 * result + (getValue() != null ? getValue().hashCode() : 0);
    result = 31 * result + (getCaseType() != null ? getCaseType().hashCode() : 0);
    return result;
  }

  @Override
  public boolean equals(Object o) {
    if (!(o instanceof SearchCriteria)) {
      return false;
    }
    SearchCriteria c2 = (SearchCriteria) o;
    if (this.getKey().equals(c2.getKey()) && this.getValue().equals(c2.getValue())
            && this.getOperationType().equals(c2.getOperationType())
            && this.isCaseInsensitive() == (c2.isCaseInsensitive())) {
      return true;
    }
    return false;
  }

  @Override
  public String toString() {
    return "SearchCriteria{" + "key='" + key + '\'' + ", operationType=" + operationType
            + ", operator='" + operation + '\'' + ", value=" + value + ", caseInsensitive="
            + caseInsensitive + '}';
  }
}
