package ch.unige.solidify.test.specification;

import static org.hamcrest.Matchers.is;
import static org.hamcrest.collection.IsIn.in;

import java.math.BigDecimal;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.OffsetDateTime;
import java.util.List;
import javax.annotation.Resource;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.boot.test.autoconfigure.orm.jpa.DataJpaTest;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;

import ch.unige.solidify.test.config.MockUserJpaConfig;
import ch.unige.solidify.test.specification.dao.MockUser;
import ch.unige.solidify.test.specification.dao.MockUserInfo;
import ch.unige.solidify.test.specification.dao.MockUserInfoRepository;
import ch.unige.solidify.test.specification.dao.MockUserRepository;
import ch.unige.solidify.test.specification.dao.MockUserSearchSpecification;
import ch.unige.solidify.util.SearchCriteria;
import ch.unige.solidify.util.SearchOperation;

@RunWith(SpringRunner.class)
@DataJpaTest
@ContextConfiguration(classes = { MockUserJpaConfig.class })
public class SearchSpecificationTest {

  @Resource
  private MockUserRepository userRepository;
  @Resource
  private MockUserInfoRepository mockUserInfoRepository;

  MockUser user1;
  MockUser user2;
  MockUser user3;
  LocalDate baseDate = LocalDate.parse("2016-08-16");
  LocalDateTime baseDateTime = LocalDateTime.parse("2015-12-03T10:15:30");
  OffsetDateTime baseOffsetTime = OffsetDateTime.parse("2007-12-03T10:15:30+01:00");

  @Before
  public void setup() {
    MockUserInfo info = new MockUserInfo("1", "John info", true);
    MockUserInfo info2 = new MockUserInfo("2", "Doe info", false);
    MockUserInfo info3 = new MockUserInfo("3", "Jo info", true);
    mockUserInfoRepository.save(info);
    mockUserInfoRepository.save(info2);
    mockUserInfoRepository.save(info3);
    mockUserInfoRepository.flush();

    user1 = new MockUser("john", 22, new Integer("3"), new Long("2000"), new BigDecimal("200.58"),
            baseDate.minusYears(10), baseDateTime.minusYears(5).minusMonths(2), baseOffsetTime);
    user2 = new MockUser("doe", 15, new Integer("1"), new Long("4000"), new BigDecimal("300.25"),
            baseDate.minusYears(20), baseDateTime.minusYears(10).minusMonths(2), baseOffsetTime);
    user3 = new MockUser("jo", 40, new Integer("5"), new Long("3000"), new BigDecimal("400.13"),
            baseDate.minusYears(30), baseDateTime.minusYears(15).minusMonths(2), baseOffsetTime);
    user1.setUserInfo(info);
    user2.setUserInfo(info2);
    user3.setUserInfo(info3);

    userRepository.save(user1);
    userRepository.save(user2);
    userRepository.save(user3);
    user1 = userRepository.findByName("john");
    user2 = userRepository.findByName("doe");
    user3 = userRepository.findByName("jo");
  }

  // Test BETWEEN with Integer field type
  @Test
  public void IntegerSearch_between_thenCorrect() {
    SearchCriteria criteria1 = new SearchCriteria("childNumber", SearchOperation.BETWEEN, "4_8");
    List<MockUser> results = userRepository.findAll(Specification.where(new MockUserSearchSpecification(criteria1)));
    Assert.assertEquals(1, results.size());
    MockUser userFound = results.get(0);
    Assert.assertEquals("jo", userFound.getName());
  }

  // Test EQUALITY with BigDecimal field type
  @Test
  public void BigDecimalSearch_equals_thenCorrect() {
    SearchCriteria criteria1 = new SearchCriteria("salary", SearchOperation.EQUALITY, "200.58");
    List<MockUser> results = userRepository.findAll(Specification.where(new MockUserSearchSpecification(criteria1)));
    Assert.assertEquals(1, results.size());
    MockUser userFound = results.get(0);
    Assert.assertEquals("john", userFound.getName());
  }

  // Test GREATER THAN OR EQUAL with Long field type
  @Test
  public void LongSearch_greateerEquals_thenCorrect() {
    SearchCriteria criteria1 = new SearchCriteria("carKilos", SearchOperation.GREATER_EQUAL, "3000");
    List<MockUser> results = userRepository.findAll(Specification.where(new MockUserSearchSpecification(criteria1)));
    Assert.assertEquals(2, results.size());
    MockUser userFound = results.get(0);
    Assert.assertEquals("doe", userFound.getName());
  }

  // Test LESS THAN OR EQUAL with Localedate field type
  @Test
  public void LocaleDateSearch_lessEquals_thenCorrect() {
    SearchCriteria criteria1 = new SearchCriteria("birthday", SearchOperation.GREATER_EQUAL, "1997-01-30");
    List<MockUser> results = userRepository.findAll(Specification.where(new MockUserSearchSpecification(criteria1)));
    Assert.assertEquals(1, results.size());
    MockUser userFound = results.get(0);
    Assert.assertEquals("john", userFound.getName());
  }

  // Test BETWEEN with LocaledateTime field type
  @Test
  public void LocaledateTimeSearch_between_thenCorrect() {
    SearchCriteria criteria1 = new SearchCriteria("marriageTime", SearchOperation.BETWEEN,
            "2003-10-03T10:15:30_2016-02-03T12:15:30");
    List<MockUser> results = userRepository.findAll(Specification.where(new MockUserSearchSpecification(criteria1)));
    Assert.assertEquals(2, results.size());
    MockUser userFound = results.get(0);
    Assert.assertEquals("john", userFound.getName());
  }

  // TEST equal with OFfsetDateTime baseOffsetTime
  @Test
  public void OffsetdateTimeSearch_equals_thenCorrect() {
    SearchCriteria criteria1 = new SearchCriteria("meetingTime", SearchOperation.EQUALITY,
            "2007-12-03T10:15:30+01:00");
    List<MockUser> results = userRepository.findAll(Specification.where(new MockUserSearchSpecification(criteria1)));
    Assert.assertEquals(3, results.size());
  }

  //
  @Test
  public void compare_2Criterias_equal() {
    SearchCriteria criteria1 = new SearchCriteria("name", SearchOperation.STARTS_WITH, "jo");
    SearchCriteria criteria2 = new SearchCriteria("name", SearchOperation.STARTS_WITH, "jo");
    Assert.assertTrue(criteria1.equals(criteria2));
    criteria2.setCaseType("i");
    Assert.assertFalse(criteria1.equals(criteria2));
    criteria2.setCaseType("c");
    criteria2.setOperation("~");
    Assert.assertFalse(criteria1.equals(criteria2));
  }

  @Test
  public void searchOne2OneEntity_ByContains_thenCorrect() {
    SearchCriteria criteria1 = new SearchCriteria("userInfo.info", SearchOperation.CONTAINS,
            "Doe");
    List<MockUser> results = userRepository.findAll(Specification.where(new MockUserSearchSpecification(criteria1)));
    Assert.assertEquals(1, results.size());
    MockUser userFound = results.get(0);
    Assert.assertEquals("Doe info", userFound.getUserInfo().getInfo());
  }

  @Test
  public void searchOne2OneEntity_ByBooleanValue_thenCorrect() {
    SearchCriteria criteria1 = new SearchCriteria("userInfo.adminuser", SearchOperation.EQUALITY,
            "true");
    List<MockUser> results = userRepository.findAll(Specification.where(new MockUserSearchSpecification(criteria1)));
    Assert.assertEquals(2, results.size());
    MockUser userFound = results.get(0);
    Assert.assertEquals(true, userFound.getUserInfo().isAdminuser());
  }

  @Test
  public void givenNameSearch_ByEqual_thenCorrect() {
    SearchCriteria criteria = new SearchCriteria("name", SearchOperation.EQUALITY, "john");
    Specification<MockUser> spec = new MockUserSearchSpecification(criteria);
    List<MockUser> results = userRepository.findAll(spec);
    Assert.assertThat(user1, is(in(results)));
  }

  /**
   * Test starts with and not
   */
  @Test
  public void givenNamesearch_ByStartsWithNot_thenCorrect() {
    SearchCriteria criteria1 = new SearchCriteria("name", SearchOperation.STARTS_WITH, "jo");
    SearchCriteria criteria2 = new SearchCriteria("name", SearchOperation.NEGATION, "jo");
    List<MockUser> results = userRepository.findAll(
            Specification.where(new MockUserSearchSpecification(criteria1))
                    .and(new MockUserSearchSpecification(criteria2)));
    MockUser userFound = results.get(0);
    Assert.assertEquals("john", userFound.getName());
    Assert.assertNotEquals("jo", userFound.getName());
  }

  /**
   * Test contains and case insensitive
   */
  @Test
  public void givenNamesearch_ByContainsCaseInsensitive_thenCorrect() {
    SearchCriteria criteria1 = new SearchCriteria("i", "name", SearchOperation.CONTAINS, "O");
    List<MockUser> results = userRepository.findAll(Specification.where(new MockUserSearchSpecification(criteria1)));
    Assert.assertEquals(3, results.size());
    SearchCriteria criteria2 = new SearchCriteria("name", SearchOperation.CONTAINS, "O");
    List<MockUser> results2 = userRepository.findAll(
            Specification.where(new MockUserSearchSpecification(criteria2)));
    Assert.assertEquals(0, results2.size());
  }

  // Test Greater Than and less Than
  @Test
  public void ageSearch_ByGreaterLess_thenCorrect() {
    SearchCriteria criteria1 = new SearchCriteria("i", "age", SearchOperation.LESS_THAN, 30);
    SearchCriteria criteria2 = new SearchCriteria("age", SearchOperation.GREATER_THAN, 20);
    List<MockUser> results = userRepository.findAll(
            Specification.where(new MockUserSearchSpecification(criteria1))
                    .and(new MockUserSearchSpecification(criteria2)));
    Assert.assertEquals(1, results.size());
    MockUser userFound = results.get(0);
    Assert.assertEquals("john", userFound.getName());
  }

}
