package ch.unige.solidify.util;

import java.io.IOException;
import java.nio.file.Paths;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.ClassPathResource;

import ch.unige.solidify.SolidifyConstants;

import gov.loc.repository.bagit.conformance.BagProfileChecker;
import gov.loc.repository.bagit.domain.Bag;
import gov.loc.repository.bagit.exceptions.CorruptChecksumException;
import gov.loc.repository.bagit.exceptions.FileNotInPayloadDirectoryException;
import gov.loc.repository.bagit.exceptions.InvalidBagitFileFormatException;
import gov.loc.repository.bagit.exceptions.MaliciousPathException;
import gov.loc.repository.bagit.exceptions.MissingBagitFileException;
import gov.loc.repository.bagit.exceptions.MissingPayloadDirectoryException;
import gov.loc.repository.bagit.exceptions.MissingPayloadManifestException;
import gov.loc.repository.bagit.exceptions.UnparsableVersionException;
import gov.loc.repository.bagit.exceptions.UnsupportedAlgorithmException;
import gov.loc.repository.bagit.exceptions.VerificationException;
import gov.loc.repository.bagit.exceptions.conformance.BagitVersionIsNotAcceptableException;
import gov.loc.repository.bagit.exceptions.conformance.FetchFileNotAllowedException;
import gov.loc.repository.bagit.exceptions.conformance.MetatdataValueIsNotAcceptableException;
import gov.loc.repository.bagit.exceptions.conformance.RequiredManifestNotPresentException;
import gov.loc.repository.bagit.exceptions.conformance.RequiredMetadataFieldNotPresentException;
import gov.loc.repository.bagit.exceptions.conformance.RequiredTagFileNotPresentException;
import gov.loc.repository.bagit.reader.BagReader;
import gov.loc.repository.bagit.verify.BagVerifier;

public class BagItTool {
  private static final Logger log = LoggerFactory.getLogger(BagItTool.class);

  /**
   * This constructor should not be called
   */
  private BagItTool() {
    throw new IllegalStateException(SolidifyConstants.TOOL_CLASS);
  }

  // Read BagIt package
  public static Bag read(String bagItPackage) {
    Bag bag = null;
    BagReader bagReader = new BagReader();
    try {
      bag = bagReader.read(Paths.get(bagItPackage));
    } catch (IOException | UnparsableVersionException | MaliciousPathException
            | UnsupportedAlgorithmException | InvalidBagitFileFormatException e) {
      log.error("Error reading BagIt package", e);
      return null;
    }
    return bag;
  }

  // Check conformance of BagIt package
  public static boolean check(String bagItProfile, String bagItPackage) {
    Bag bag = read(bagItPackage);
    if (bag == null) {
      log.error("Cannot read BagIt{}", bagItPackage);
      return false;
    }
    boolean ignoreHiddenFiles = true;
    ExecutorService exeService = new ThreadPoolExecutor(0, 3000, 60L, TimeUnit.SECONDS, new LinkedBlockingQueue<Runnable>());
    try (BagVerifier bagVerifier = new BagVerifier(exeService)) {
      bagVerifier.isValid(bag, ignoreHiddenFiles);
      BagProfileChecker.bagConformsToProfile(new ClassPathResource(bagItProfile).getInputStream(),
              bag);
    } catch (IOException | MissingPayloadManifestException | MissingBagitFileException
            | MissingPayloadDirectoryException | FileNotInPayloadDirectoryException
            | InterruptedException | MaliciousPathException | UnsupportedAlgorithmException
            | InvalidBagitFileFormatException | CorruptChecksumException | VerificationException
            | FetchFileNotAllowedException | RequiredMetadataFieldNotPresentException
            | MetatdataValueIsNotAcceptableException | RequiredManifestNotPresentException
            | BagitVersionIsNotAcceptableException | RequiredTagFileNotPresentException e) {
      Thread.currentThread().interrupt();
      log.error("Error while checking BagIt package", e);
      return false;
    }
    return true;
  }

}
