package ch.unige.solidify.util;

import java.io.UnsupportedEncodingException;
import java.lang.reflect.Array;
import java.net.URI;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.ArrayList;
import java.util.Base64;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.UUID;
import javax.xml.bind.DatatypeConverter;

import org.json.JSONArray;
import org.json.JSONObject;
import org.json.JSONString;
import org.json.XML;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.ObjectMapper;

import ch.unige.solidify.SolidifyConstants;

public class StringTool {
  private static final Logger log = LoggerFactory.getLogger(StringTool.class);
  public static final String DATE_TIME_FORMAT = "yyyy-MM-dd'T'HH:mm:ss.SSSZ";
  public static final String DATE_FORMAT = "yyyy-MM-dd";

  /**
   * This constructor should not be called
   */
  private StringTool() {
    throw new IllegalStateException(SolidifyConstants.TOOL_CLASS);
  }

  public static String generateResId() {
    return UUID.randomUUID().toString();
  }

  public static boolean toUpdate(String vOri, String vNew) {
    if (vNew == null || vNew.isEmpty())
      return false;
    return (vOri == null || !vOri.equals(vNew));
  }

  @SuppressWarnings("squid:S00117")
  public static String XML2JSON(String xmlString) {
    JSONObject json = XML.toJSONObject(xmlString);
    return json.toString();
  }

  @SuppressWarnings("squid:S00117")
  public static String JSON2XML(String jsonString) {
    JSONObject json = new JSONObject(jsonString);
    return XML.toString(json);
  }

  @SuppressWarnings("squid:S00117")
  public static String Map2JSON(Map<String, Object> metadata) {
    JSONObject json = new JSONObject(metadata);
    return json.toString();
  }

  @SuppressWarnings("squid:S00117")
  public static String Map2XML(Map<String, Object> metadata) {
    return JSON2XML(Map2JSON(metadata));
  }

  public static void filterMap(Map<String, Object> metadata, String[] exceptions) {
    filterObject(metadata, exceptions);
  }

  private static void filterObject(Object obj, String[] exceptions) {
    if (obj == null) {
      return;
    }
    if (obj instanceof JSONObject || obj instanceof JSONArray || obj instanceof JSONString
            || obj instanceof Byte || obj instanceof Character || obj instanceof Short
            || obj instanceof Integer || obj instanceof Long || obj instanceof Boolean
            || obj instanceof Float || obj instanceof Double || obj instanceof String) {
      return;
    }

    if (obj instanceof Collection) {
      Iterator<?> i = ((Collection<?>) obj).iterator();
      while (i.hasNext()) {
        filterObject(i.next(), exceptions);
      }
    }
    if (obj.getClass().isArray()) {
      int length = Array.getLength(obj);
      for (int i = 0; i < length; i += 1) {
        filterObject(Array.get(obj, i), exceptions);
      }
    }
    if (obj instanceof Map) {
      List<String> toDelete = new ArrayList<>();
      Iterator<?> i = ((Map<?, ?>) obj).entrySet().iterator();
      while (i.hasNext()) {
        Map.Entry<?, ?> e = (Map.Entry<?, ?>) i.next();
        String k = ((String) e.getKey());
        // Do not index exception
        for (String x7 : exceptions) {
          if (k.endsWith(x7)) {
            toDelete.add(k);
            break;
          }
        }
        // Remove empty object
        if (e.getValue() instanceof String) {
          String v = ((String) e.getValue());
          if (StringTool.isNullOrEmpty(v)) {
            log.trace("To remove {}", k);
            toDelete.add(k);
          }
        } else {
          filterObject(e.getValue(), exceptions);
        }
      }
      for (String k : toDelete) {
        log.trace("Removing {}", k);
        ((Map<?, ?>) obj).remove(k);
      }
    }
  }

  public static String formatDate(OffsetDateTime date, DateTimeFormatter formatTemplate) {
    return date.toLocalDateTime().format(formatTemplate);
  }

  /**
   * Check if the date is valid
   * 
   * @param dateString
   *          the date as a string
   * @return true if valid false if not
   */
  public static boolean checkDate(String dateString) {
    if (dateString.matches("\\d{4}\\-\\d{2}\\-\\d{2}")) {
      dateString += "T00:00:00Z";
    }
    try {
      getDate(dateString);
      return true;
    } catch (DateTimeParseException e) {
      return false;
    }
  }

  /**
   * Get an OffsetDateTime instance from a String
   * 
   * @param dateString
   *          the date as a string
   * @return the date as an OffsetDateTime
   * @throws DateTimeParseException
   */
  public static OffsetDateTime getDate(String dateString) {
    return OffsetDateTime.parse(dateString).toInstant().atOffset(ZoneOffset.UTC);
  }

  /*
   * Get an OffsetDateTime instance from a String and a DateTimeFormatter
   * 
   * @param dateString the date as a string
   * 
   * @param formatter the formatter
   * 
   * @return the date as an OffsetDateTime
   * 
   * @throws DateTimeParseException
   */
  public static OffsetDateTime getDate(String dateString, DateTimeFormatter formatter) {
    return OffsetDateTime.parse(dateString, formatter).toInstant().atOffset(ZoneOffset.UTC);
  }

  public static String getUrlPath(String url) {
    URI uri = URI.create(url.trim());
    return uri.getPath();
  }

  public static String encodeUrl(String s) throws UnsupportedEncodingException {
    return URLEncoder.encode(s, StandardCharsets.UTF_8.toString());
  }

  public static String decodeUrl(String s) throws UnsupportedEncodingException {
    return URLDecoder.decode(s, StandardCharsets.UTF_8.toString());
  }

  public static String formatSmartSize(long size) {
    String unit;
    double s;
    if (size < 1024l) {
      s = size;
      unit = "B";
    } else if (size < (1024l * 1024l)) {
      s = (double) size / 1024l;
      unit = "KB";
    } else if (size < (1024l * 1024l * 1024l)) {
      s = (double) size / (1024l * 1024l);
      unit = "MB";
    } else if (size < (1024l * 1024l * 1024l * 1024l)) {
      s = (double) size / (1024l * 1024l * 1024l);
      unit = "GB";
    } else {
      s = (double) size / (1024l * 1024l * 1024l * 1024l);
      unit = "TB";
    }
    return String.format(Locale.ROOT, "%.1f %s", s, unit);
  }

  public static boolean isNullOrEmpty(String value) {
    return (value == null || value.isEmpty());
  }

  public static String toCamelCase(String strTocamel) {
    if (isNullOrEmpty(strTocamel)) {
      return strTocamel;
    }
    return strTocamel.substring(0, 1).toUpperCase() + strTocamel.substring(1).toLowerCase();
  }

  public static String capitalize(String str) {
    return StringTool.capitalize(str, true);
  }

  public static String capitalize(String str, boolean lowercase) {
    if (!StringTool.isNullOrEmpty(str)) {
      if (lowercase) {
        return str.substring(0, 1).toUpperCase() + str.substring(1).toLowerCase();
      } else {
        return str.substring(0, 1).toUpperCase() + str.substring(1);
      }
    } else {
      return str;
    }
  }

  public static String encode64(String string) {
    return Base64.getEncoder().withoutPadding().encodeToString(string.getBytes());
  }

  public static String encode64WithPadding(String string) {
    return Base64.getEncoder().encodeToString(string.getBytes());
  }

  public static String decode64(String encodedstring) {
    return new String(Base64.getDecoder().decode(encodedstring));
  }

  public static String decode64IfEncoded(String string) {
    if (string.startsWith(SolidifyConstants.BASE64_PREFIX)) {
      return StringTool.decode64(string.substring(SolidifyConstants.BASE64_PREFIX.length()));
    }
    return string;
  }

  public static String hash(String algo, String s) throws NoSuchAlgorithmException {
    MessageDigest md = MessageDigest.getInstance(algo);
    md.update(s.getBytes());
    byte[] digest = md.digest();
    return DatatypeConverter.printHexBinary(digest);
  }

  public static String toHexaString(byte[] bytes) {
    // Convert it to hexadecimal format
    StringBuilder sb = new StringBuilder();
    for (byte aByte : bytes) {
      sb.append(Integer.toString((aByte & 0xff) + 0x100, 16).substring(1));
    }
    return sb.toString();
  }

  public static String truncate(String str, int length) {
    if (StringTool.isNullOrEmpty(str) || length >= str.length() || length < 0) {
      return str;
    }
    return str.substring(0, length);
  }

  public static String truncateWithEllipsis(String str, int length) {
    return StringTool.truncateWithEllipsis(str, length, "...");
  }

  public static String truncateWithEllipsis(String str, int length, String ellipsis) {
    if (StringTool.isNullOrEmpty(str) || length >= str.length() || length < 0
            || StringTool.isNullOrEmpty(ellipsis)) {
      return str;
    } else {
      int lengthWithoutEllipsis = length - ellipsis.length();
      if (lengthWithoutEllipsis > 0) {
        return StringTool.truncate(str, lengthWithoutEllipsis) + ellipsis;
      } else {
        return StringTool.truncate(str, length);
      }
    }
  }

  public static String convertToQueryString(Object object) {

    // Object --> map
    ObjectMapper objectMapper = new ObjectMapper();
    @SuppressWarnings("unchecked")
    Map<String, Object> map = objectMapper.convertValue(object, Map.class);

    StringBuilder queryString = new StringBuilder();
    for (String key : map.keySet()) {

      // Do not process empty object or HATEOAS links
      if (map.get(key) == null || key.equals("links")) {
        continue;
      }
      // Add & if needed
      if (queryString.length() > 0) {
        queryString.append("&");
      }
      // key=value
      queryString.append(key);
      queryString.append("=");
      queryString.append(map.get(key));
    }
    return queryString.toString();
  }

  /**
   * Convert path with Windows File Separator (backslash) to path with Unix File Separator (slash)
   * 
   * @param path
   *          the path to convert in unix path
   * @return path with slash as file separator
   */
  public static String convertInWindowsPath(String path) {
    if (path == null) {
      return null;
    }
    return path.replaceAll("\\\\", SolidifyConstants.URL_SEP);
  }
}
