package ch.unige.solidify.util;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.unige.solidify.SolidifyConstants;

public class ZipTool {
  private static final Logger log = LoggerFactory.getLogger(ZipTool.class);
  private static final int BUFFER_SIZE = 4096;

  private final Path archive;

  public ZipTool(URI archive) {
    this.archive = Paths.get(archive);
  }

  public ZipTool(String archive) {
    this.archive = Paths.get(archive);
  }

  public String getArchive() {
    return archive.toString();
  }

  public boolean zipFiles(URI source) {
    return zipFiles(Paths.get(source));
  }

  public boolean zipFiles(Path source) {
    byte[] buffer = new byte[BUFFER_SIZE];
    int count = 0;
    try (ZipOutputStream zipOut = new ZipOutputStream(new FileOutputStream(this.getArchive()))) {
      for (Path file : generateFileList(source)) {
        count++;
        Path f = source.resolve(file);
        ZipEntry zipEntry = new ZipEntry(normalizePath(file.toString()));
        zipOut.putNextEntry(zipEntry);
        if (FileTool.isFile(f)) {
          try (FileInputStream fileIn = new FileInputStream(f.toFile())) {
            int len;
            while ((len = fileIn.read(buffer)) > 0) {
              zipOut.write(buffer, 0, len);
            }
          }
        }
        zipOut.closeEntry();
      }
    } catch (FileNotFoundException e) {
      log.error("Cannot create Zip package (" + archive + ")", e);
      return false;
    } catch (IOException e) {
      log.error("Problem in creating Zip package (" + source + ".zip)", e);
      return false;
    }
    log.info("Zip package ('" + source + "' in '" + this.archive + "') created with " + count
            + " entries");
    return true;
  }

  public boolean unzipFiles(URI destination) {
    return unzipFiles(Paths.get(destination));
  }

  public boolean unzipFiles(Path destination) {
    int count = 0;
    try (ZipInputStream zis = new ZipInputStream(new FileInputStream(this.getArchive()))) {
      ZipEntry ze = zis.getNextEntry();
      while (ze != null) {
        extractFile(zis, ze, destination, null);
        count++;
        ze = zis.getNextEntry();
      }
      zis.closeEntry();
    } catch (IOException e) {
      log.error("Problem in unzipping package (" + this.archive + ")", e);
      return false;
    }
    log.info("Unzip package ('" + this.archive + "') in '" + destination + "': " + count
            + " entries extracted");
    return true;
  }

  public boolean unzipFolder(Path destination, String folder) {
    return unzipFiles(destination, Collections.singletonList(folder), folder);
  }

  public boolean unzipFiles(Path destination, List<String> list) {
    return unzipFiles(destination, list, null);
  }

  private boolean unzipFiles(Path destination, List<String> list, String relativeFolder) {
    try (InputStream is = new FileInputStream(this.getArchive())) {
      return unzipFiles(is, destination, list, relativeFolder);
    } catch (IOException e) {
      log.error("Problem in unzipping package (" + this.archive + ")", e);
    }
    return false;
  }

  public static boolean unzipFiles(InputStream zipData, Path destination, List<String> list,
          String relativeFolder) {
    int count = 0;
    try (ZipInputStream zis = new ZipInputStream(zipData)) {
      ZipEntry ze = zis.getNextEntry();
      while (ze != null) {
        if (isFileCandidate(list, ze)) {
          extractFile(zis, ze, destination, relativeFolder);
          count++;
        }
        ze = zis.getNextEntry();
      }
      zis.closeEntry();

    } catch (IOException e) {
      log.error("Problem in unzipping package " + destination, e);
      return false;
    }
    log.info("Unzip package in '" + destination + "': " + count + " entries extracted");
    return true;

  }

  private static boolean isFileCandidate(List<String> list, ZipEntry ze) {
    if (list == null) {
      return true;
    }
    for (String f : list) {
      String zePath = ze.getName();
      if (zePath.startsWith(normalizePath(f))) {
        return true;
      }
    }
    return false;
  }

  private static void extractFile(ZipInputStream zipFile, ZipEntry zipEntry, Path destination,
          String relativeFolder) throws IOException {
    Path newFile = checkDestination(zipEntry, destination, relativeFolder);
    if (newFile != null) {
      try (FileOutputStream fos = new FileOutputStream(newFile.toFile())) {
        byte[] buffer = new byte[BUFFER_SIZE];
        int len;
        while ((len = zipFile.read(buffer)) > 0) {
          fos.write(buffer, 0, len);
        }
      } catch (IOException e) {
        log.error("IOException when extracting zip file " + zipFile, e);
        throw e;
      }
    }
  }

  private static Path checkDestination(ZipEntry zipEntry, Path destination, String relativeFolder)
          throws IOException {
    Path newFile;
    if (relativeFolder == null) {
      newFile = destination.resolve(zipEntry.getName()).toAbsolutePath();
    } else {
      String newName = zipEntry.getName()
              .substring(zipEntry.getName().indexOf(normalizePath(relativeFolder)) + relativeFolder.length());
      newFile = destination.resolve(newName).toAbsolutePath();
    }
    if (zipEntry.isDirectory()) {
      FileTool.ensureFolderExists(newFile);
      return null;
    } else {
      if (!FileTool.ensureFolderExists(newFile.getParent())) {
        String message = "Cannot create parent directory (" + newFile + ")";
        log.error(message);
        throw new IOException(message);
      }
      return newFile;
    }
  }

  public int getEntryNumber() {
    try (ZipFile zipFile = new ZipFile(this.getArchive())) {
      return zipFile.size();
    } catch (IOException e) {
      log.error("IOException when getting entry number", e);
      return 0;
    }
  }

  public boolean contains(String fileName) {
    try (ZipInputStream zis = new ZipInputStream(new FileInputStream(this.getArchive()))) {
      ZipEntry entry;
      while ((entry = zis.getNextEntry()) != null) {
        if (entry.getName().equals(fileName)) {
          return true;
        }
      }
    } catch (IOException e) {
      log.error("Problem in reading package (" + this.archive + ")", e);
    }
    return false;
  }

  public boolean checkZip() {
    try (ZipFile zipFile = new ZipFile(this.getArchive())) {
      Enumeration<? extends ZipEntry> entries = zipFile.entries();
      while (entries.hasMoreElements()) {
        ZipEntry zipEntry = entries.nextElement();
        zipFile.getInputStream(zipEntry);
        zipEntry.getCrc();
        zipEntry.getCompressedSize();
        zipEntry.getName();
      }
      return true;
    } catch (IOException e) {
      log.error("IOException when checking zip file " + this.getArchive(), e);
      return false;
    }
  }

  private List<Path> generateFileList(Path src) throws IOException {
    try (Stream<Path> walk = Files.walk(src)) {
      return walk.filter(path -> path.toFile().isFile()).map(path -> src.relativize(path))
              .collect(Collectors.toList());
    }
  }

  private static String normalizePath(String path) {
    return path.replaceAll("\\\\", SolidifyConstants.URL_SEP);
  }

}
